@@ -6,6 +6,9 @@
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
+# This is created in the caller's space
+BEGIN
+  { sub ::PCDATA { '#PCDATA' } }
 
 ######################################################################
 package XML::Twig;
@@ -16,16 +19,16 @@ require 5.004;
 use strict; 
 use vars qw($VERSION @ISA);
 use Carp;
-#use Exporter;
+
 
 sub PCDATA { '#PCDATA' }
 
 BEGIN
 {
     require XML::Parser;
-    $VERSION = '1.6';
+    $VERSION = '1.7';
 
-    my $needVersion = '2.16';
+    my $needVersion = '2.21';
     croak "need at least XML::Parser version $needVersion"
 	unless $XML::Parser::VERSION >= $needVersion;
     import XML::Twig::Elt;
@@ -35,12 +38,25 @@ BEGIN
 
 @ISA = ("XML::Parser");
 
-my %twig_handlers=( Start  => \&twig_start, 
-                    End    => \&twig_end, 
-                    Char   => \&twig_char, 
-                    Entity => \&twig_entity, 
+my %twig_handlers=( Start   => \&twig_start, 
+                    End     => \&twig_end, 
+                    Char    => \&twig_char, 
+                    Entity  => \&twig_entity, 
+                    XMLDecl => \&twig_xmldecl, 
+                    Doctype => \&twig_doctype, 
+                    Element => \&twig_element, 
+                    Attlist => \&twig_attlist, 
                   );
 
+# those handlers are only used if DTD info is needed
+#my %twig_dtd_handlers= ( Element => \&twig_element, 
+#                         Attlist => \&twig_attlist, 
+#                       );
+
+# those handlers are only used if the entities are NOT to be expanded
+my %twig_noexpand_handlers= ( Default => \&twig_default, 
+                            );
+
 my $ID= 'id'; # default value, set by the Id argument
 
 # used to store the gi's
@@ -48,7 +64,16 @@ my $ID= 'id'; # default value, set by the Id argument
 my %gi2index; # gi => index
 my @index2gi; # list of gi's
 
-my $XML_DECL= '<?xml version="1.0" standalone="%s"?>';
+my $XML_DECL= '<?xml version="1.0" ?>';
+
+# used to replace the predefined entities if NoExpand is used
+my %base_ent= ( '>' => '&gt;',
+                '<' => '&lt;',
+                '&' => '&amp;',
+                "'" => '&apos;',
+                '"' => '&quot;',
+              );
+
 
 1;
 
@@ -60,18 +85,35 @@ sub new
       { $handlers= $args{TwigHandlers};
         delete $args{TwigHandlers};
       }
-    
-    if( exists $args{Id})
+
+   if( exists $args{Id})
       { $ID= $args{Id};
         delete $args{ID};
       }
     
     my $self= new XML::Parser(%args, 'Handlers' => \%twig_handlers);
     $self->{twig_handlers}= $handlers;
+
+     if( exists $args{NoExpand})
+      { $self->setHandlers( %twig_noexpand_handlers);
+      }
+
+    # take care of twig-specific arguments
+    $self->{twig_starttag_handlers}= $args{StartTagHandlers} || {};
+    delete $args{StartTagHandlers};
+    $self->{twig_dtd_handler}= $args{DTDHandler};
+    delete $args{DTDHandler};
+    $self->{twig_char_handler}= $args{CharHandler};
+    delete $args{CharHandler};
+    $self->{twig_read_external_dtd}= $args{LoadDTD};
+    delete $args{LoadDTD};
     $self->{twig}= $self;
 
+    
     $self->{twig_entity_list}= new XML::Twig::Entity_list; 
 
+    $self->{twig_id}= $ID; 
+
     bless $self, $class;
     return $self;
 
@@ -81,8 +123,12 @@ sub new
 sub twig_start($$%)
   { my ($p, $gi, %att)  = @_;
     my $t=$p->{twig};
-    my $parent= $t->{twig_elt};
+    my $parent= $t->{twig_current};
     my $elt= new XML::Twig::Elt();
+ 
+    delete $parent->{twig_current} if( $parent);
+    $t->{twig_current}= $elt;
+    $elt->{twig_current}=1;
 
     $elt->set_gi( $gi);
     $elt->set_atts( \%att);
@@ -102,6 +148,10 @@ sub twig_start($$%)
       { # processing root
         $t->{twig_root}= $elt;
         $elt->{twig}= $t;
+        # call dtd handlerif need be
+        $t->{twig_dtd_handler}->($t, $t->{twig_dtd})
+          if( defined $t->{twig_dtd_handler});
+          
       }
     
     if( $p->recognized_string=~ /\/>\Z/) { $elt->{empty}=1; }
@@ -110,11 +160,14 @@ sub twig_start($$%)
     my $id= $elt->id;
     if( $id) { $t->{twig_id_list}->{$id}= $elt; }
 
-    $t->{twig_elt}= $elt;
 
     # empty the stored pcdata (space stored in case
     # they are really part of a pcdata element)
     $t->{twig_stored_pcdata}='';
+
+    # call user handler if need be
+    if( defined $t->{twig_starttag_handlers}->{$gi})
+      { $t->{twig_starttag_handlers}->{$gi}->($t, $elt); }
   }
 
 
@@ -123,11 +176,13 @@ sub twig_end($$%)
     my $t=$p->{twig};
         
     # build the tree
-    my $elt= $t->{twig_elt};
-    if( $elt) { $t->{twig_parent}= $elt->parent; }
-    $t->{twig_elt}= $t->{twig_elt}->parent;
+    delete $t->{twig_current}->{twig_current} if( $t->{twig_current});
+    my $elt= $t->{twig_current};
+    $elt->{twig_current}=1;
 
-    # empty the stored pcdata (space stored in case
+    $t->{twig_current}= $t->{twig_current}->parent;
+
+    # empty the stored pcdata (spaces stored in case
     # they are really part of a pcdata element)
     $t->{twig_stored_pcdata}='';
 
@@ -137,12 +192,23 @@ sub twig_end($$%)
 
   }
 
-sub twig_char($$)
-  { my ($p, $string)= @_;
+sub twig_char($$$)
+  { my ($p, $string, $called_from_default)= @_;
     my $t=$p->{twig};
 
+    if( $t->{twig_char_handler})
+      { $string= $t->{twig_char_handler}->( $string); }
+
+    # why o why does Expat silently convert XML base enttities?
+    if( !$called_from_default)
+      { $string=~ s/&/&amp;/g; }
+    $string=~ s/([<>'"])/$base_ent{$1}/g;
+
+
+    delete $t->{twig_current}->{twig_current} if( $t->{twig_current});
+    my $elt= $t->{twig_current};
+    $elt->{twig_current}=1;
 
-    my $elt= $t->{twig_elt};
     if( $elt->gi eq PCDATA)
       { # text is the continuation of a previously created pcdata
         $elt->set_pcdata( $elt->{pcdata}.$string); } 
@@ -153,11 +219,12 @@ sub twig_char($$)
             return;
           } 
         # create a new #PCDATA element
-        my $parent= $t->{twig_elt};    # always defined
+        my $parent= $t->{twig_current};    # always defined
         my $elt=  new XML::Twig::Elt();
         $elt->set_gi( PCDATA);
         # if empty spaces had been previously stored they have to be added
         $string= $t->{twig_stored_pcdata}.$string;
+        $t->{twig_stored_pcdata}='';
         $elt->set_pcdata( $string);
         my $prev_sibling= $parent->last_child;
         if( $prev_sibling) 
@@ -169,55 +236,220 @@ sub twig_char($$)
         $parent->set_first_child( $elt) unless( $parent->first_child); 
         $parent->set_last_child( $elt);
       }
+    delete $t->{twig_current}->{twig_current} if( $t->{twig_current});
+    $elt->{twig_current}=1;
+   $t->{twig_current}= $elt;
   }
  
 
 sub twig_entity($$$$$$)
-  { my( $t, $name, $val, $sysid, $pubid, $ndata)= @_;
+  { my( $p, $name, $val, $sysid, $pubid, $ndata)= @_;
+    my $t=$p->{twig};
     my $ent=new XML::Twig::Entity( $name, $val, $sysid, $pubid, $ndata);
     $t->{twig_entity_list}->add( $ent);
   }
 
+sub twig_xmldecl
+  { my $p= shift;
+    my $t=$p->{twig};
+    $t->{twig_xmldecl}={};
+    $t->{twig_xmldecl}->{version}= shift;
+    $t->{twig_xmldecl}->{encoding}= shift; 
+    $t->{twig_xmldecl}->{standalone}= shift;
+  }
+
+sub twig_doctype
+  { my( $p, $name, $sysid, $pub, $internal)= @_;
+    my $t=$p->{twig};
+    $t->{twig_doctype}= {};                    # create 
+    $t->{twig_doctype}->{name}= $name;         # always there
+    $t->{twig_doctype}->{sysid}= $sysid;       #  
+    $t->{twig_doctype}->{pub}= $pub;           #  
+    $t->{twig_doctype}->{internal}=$internal; 
+    # now check if we want to get the DTD info
+    if( $t->{twig_read_external_dtd} && $sysid)
+      { # let's build a fake document with an internal DTD
+        # is this portable?
+        my $tmpfile= "twig_tmp$$";
+        open( TMP, ">$tmpfile") 
+          or die "cannot create temp file $tmpfile: $!";
+        print TMP "<!DOCTYPE $name [\n";   # print the doctype
+        # slurp the DTD
+          { open( DTD, "<$sysid") 
+              or die "cannot open dtd file $sysid: $!";
+            local undef $/;
+            my $dtd= <DTD>;
+            close DTD;
+            print TMP $dtd;                 # add the dtd
+          }
+        print TMP "]>";                     # close the dtd
+        print TMP "<dummy></dummy>\n";      # XML::Parser needs an element
+
+        close TMP;
+        
+        my $t_dtd= new XML::Twig;           # create a temp twig
+        $t_dtd->parsefile( $tmpfile);       # parse it
+        $t->{twig_dtd}= $t_dtd->{twig_dtd}; # grab the dtd info
+        unlink $tmpfile;
+      }
+
+  }
+
+sub twig_element
+  { my( $p, $name, $model)= @_;
+    my $t=$p->{twig};
+    $t->{twig_dtd}||= {};                     # may create the dtd 
+    $t->{twig_dtd}->{model}||= {};            # may create the model hash 
+    $t->{twig_dtd}->{elt_list}||= [];         # ordered list of elements 
+    push @{$t->{twig_dtd}->{elt_list}}, $name; # store the elt
+    $t->{twig_dtd}->{model}->{$name}= $model; # store the model
+  }
+
+sub twig_attlist
+  { my( $p, $el, $att, $type, $default, $fixed)= @_;
+    my $t=$p->{twig};
+    $t->{twig_dtd}||= {};                      # create dtd if need be 
+    $t->{twig_dtd}->{$el}||= {};               # create elt if need be 
+    $t->{twig_dtd}->{$el}->{att}||= {};        # create att if need be 
+    $t->{twig_dtd}->{att}->{$el}->{$att}= {} ;
+    $t->{twig_dtd}->{att}->{$el}->{$att}->{type}= $type; 
+    $t->{twig_dtd}->{att}->{$el}->{$att}->{default}= $default; 
+    $t->{twig_dtd}->{att}->{$el}->{$att}->{fixed}= $fixed; 
+  }
+
+# called only when the NoExpand option is set
+sub twig_default
+  { my( $p, $string)= @_;
+
+    # process only if we have an entity
+    return unless( $string=~ /\A&[^;]*;/);
+    
+    # the entity has to be in a pcdata, or we have a problem
+    # the 1 is used by twig_char to avoid replacing the &
+    twig_char( $p, $string, 1);
+  }
+    
+
+# returns the declaration text (or a default one)
+sub xmldecl
+  { my $t= shift;
+    my $decl= $t->{'twig_xmldecl'} or return $XML_DECL;
+    my $string= "<?xml";
+    $string .= " version=\"$decl->{version}\""       if( $decl->{version});
+    $string .= " encoding=\"$decl->{encoding}\""     if( $decl->{encoding});
+    $string .= " standalone=\"";  
+    $string .= $decl->{standalone} ? "yes" : "no";  
+    $string .= "\"";  
+    $string .= "?>\n";
+    return $string;
+  }
+
+# returns the doctype text (or none)
+# that's the doctype just has it was in the original document
+sub doctype
+  { my $t= shift;
+    my $doctype= $t->{'twig_doctype'} or return '';
+    my $string= "<!DOCTYPE ".$doctype->{name};
+    $string .= " SYSTEM \"$doctype->{sysid}\"" if( $doctype->{sysid});
+    $string .= " PUBLIC  \"$doctype->{pub}\""  if( $doctype->{pub});
+    $string .=  $doctype->{internal}.">\n";
+    return $string;
+  }
+
+# return the dtd object
+sub dtd
+  { my $t= shift;
+    return $t->{twig_dtd};
+  }
+
+# return an element model, or the list of element models
+sub model
+  { my $t= shift;
+    my $elt= shift;
+    return $t->dtd->{'model'}->{$elt} if( $elt);
+    return sort keys %{$t->{'dtd'}->{'model'}};
+  }
+
+        
+# return the entity_list object 
 sub entity_list($)
   { my $t= shift;
     return $t->{twig_entity_list};
   }
 
+# return the list of entity names 
+sub entity_names($)
+  { my $t= shift;
+    return sort keys %{$t->{twig_entity_list}} ;
+  }
+
+# return the entity object 
+sub entity($$)
+  { my $t= shift;
+    my $entity_name= shift;
+    return $t->{twig_entity_list}->{$entity_name};
+  }
+
 
 sub print_prolog
-  { my( $t, $fh)= @_;
-    $fh ||= *STDOUT;
+  { my $t= shift;
+    my $fh= ref $_[0] eq 'GLOB' ? shift : \*STDOUT;
+    my %args= @_;
+
+    my $update_dtd = $args{Update_DTD} || '';
+
     croak "no twig root" unless( $t->root);
-    if( $t->{twig_entity_list})
-      { printf $fh $XML_DECL, 'no';
-      
-        print $fh "<!DOCTYPE " . $t->root->gi . "[\n";
-        $t->{twig_entity_list}->print( $fh);
-        print "]>\n";
+    printf $fh $t->xmldecl;
+    return unless( defined $t->{'twig_doctype'});
+    my $doctype= $t->{'twig_doctype'};
+    if( $update_dtd)
+      { 
+        if( defined $doctype->{sysid}  )  
+          { print $fh "<!DOCTYPE ".$doctype->{name};
+            print $fh " SYSTEM \"$doctype->{sysid}\"";
+            print $fh " PUBLIC  \"$doctype->{pub}\""  if( $doctype->{pub});
+            print $fh "[\n";
+            print $fh $t->{twig_entity_list}->print( $fh);
+            print $fh "]>\n";
+          }
+        else
+          { my $dtd= $t->{'twig_dtd'};
+            print $fh $t->dtd_text;
+          }            
       }
     else
-      { printf $fh $XML_DECL, 'yes'; }
+      { print $fh "<!DOCTYPE ".$doctype->{name} if( $doctype->{name});
+        print $fh $doctype->{internal}; 
+        print $fh ">\n"; 
+      }
   }
 
 sub print
-  { my( $t, $fh)= @_;
-    $fh ||= *STDOUT;
-    $t->flush( $fh);
+  { my $t= shift;
+    my $fh= ref $_[0] eq 'GLOB' ? shift : \*STDOUT;
+    my %args= @_;
+
+    my $update_dtd = $args{Update_DTD} || '';
+    $t->print_prolog( $fh, Update_DTD => $update_dtd);
+    $t->{twig_root}->print( $fh);
   }
 
 sub flush
-  { my( $t, $fh)= @_;
-    $fh ||= *STDOUT;
+  { my $t= shift;
+    my $fh= ref $_[0] eq 'GLOB' ? shift : \*STDOUT;
+    my %args= @_;
+
+    my $update_dtd = $args{Update_DTD} || '';
 
     # the "real" last element processed, as twig_end has closed it
     my $last_elt;
-    if( $t->{twig_elt})
-      { $last_elt= $t->{twig_elt}->last_child; }
+    if( $t->{twig_current})
+      { $last_elt= $t->{twig_current}->last_child; }
     else
       { $last_elt= $t->{twig_root}; }
     
     my $elt= $t->{twig_root};
-    $t->print_prolog( $fh) unless( $elt->flushed);
+    $t->print_prolog( $fh, Update_DTD => $update_dtd) unless( $elt->flushed);
 
     while( $elt)
       { my $next_elt; 
@@ -233,7 +465,7 @@ sub flush
         else
           { # an element before the last one or the last one,
             $next_elt= $elt->next_sibling;  
-            $elt->flush;
+            $elt->flush( $fh);
             $elt->delete; 
             last if( $elt == $last_elt);
           }
@@ -259,6 +491,41 @@ sub change_gi
     $gi2index{$new_gi}= $index;
   }
 
+
+# builds the DTD from the stored (possibly updated) data
+sub dtd_text
+  { my $t= shift;
+    my $dtd= $t->{twig_dtd};
+    my $doctype= $t->{'twig_doctype'} or return '';
+    my $string= "<!DOCTYPE ".$doctype->{name}."[\n";
+    foreach my $gi (@{$dtd->{elt_list}})
+      { $string.= "<!ELEMENT $gi ".$dtd->{'model'}->{$gi}.">\n" ;
+        if( $dtd->{att}->{$gi})
+          { my $attlist= $dtd->{att}->{$gi};
+            $string.= "<!ATTLIST $gi\n";
+            foreach my $att ( sort keys %{$attlist})
+              { $string.= "   $att $attlist->{$att}->{type} ".
+                            "$attlist->{$att}->{default}"; 
+                if( $attlist->{$att}->{fixed})
+                  { $string .= " #FIXED"};
+                $string.= "\n";
+              }
+            $string.= ">\n";
+          }
+      }
+    $string.= $t->entity_list->text if( $t->entity_list);
+    $string.= "\n]>\n";
+    return $string;
+  }
+        
+# prints the DTD from the stored (possibly updated) data
+sub dtd_print
+  { my $t= shift;
+    my $fh= shift || *STDOUT;
+    print $fh, $t->dtd_text;
+  }
+        
+
 ######################################################################
 package XML::Twig::Entity_list;
 ######################################################################
@@ -289,7 +556,6 @@ sub delete
         my $name= shift;
         delete $list->{$name};
       }
-
   }
 
 sub print
@@ -299,6 +565,16 @@ sub print
       { $ent_list->{$ent}->print( $fh); }
   }
 
+sub text
+  { my ($ent_list)= @_;
+    return join "\n", map { $ent_list->{$_}->text} sort keys %{$ent_list};
+  }
+
+sub list
+  { my ($ent_list)= @_;
+    return @{[$ent_list]};
+  }
+
 ######################################################################
 package XML::Twig::Entity;
 ######################################################################
@@ -320,6 +596,11 @@ sub new
     return $self;
   }
 
+sub name  { return $_[0]->{name}; }
+sub val   { return $_[0]->{val}; }
+sub sysid { return $_[0]->{sysid}; }
+sub pubid { return $_[0]->{pubid}; }
+sub ndata { return $_[0]->{ndata}; }
 
 sub print
   { my ($ent, $fh)= @_;
@@ -329,16 +610,16 @@ sub print
 
 sub text
   { my ($ent)= @_;
-    if( $ent->{val})
-      { if( $ent->{val}=~ /"/)
-          { return "<!ENTITY $ent->{name} '$ent->{val}'>"; }
-        return "<!ENTITY $ent->{name} \"$ent->{val}\">";
+    if( $ent->{'val'})
+      { if( $ent->{'val'}=~ /"/)
+          { return "<!ENTITY $ent->{'name'} '$ent->{'val'}'>"; }
+        return "<!ENTITY $ent->{'name'} \"$ent->{'val'}\">";
       }
-    elsif( $ent->{sysid})
-      { my $text= "<!ENTITY $ent->{name} ";
-        $text .= "SYSTEM \"$ent->{sysid}\" " if( $ent->{sysid});
-        $text .= "PUBLIC \"$ent->{pubid}\" " if( $ent->{pubid});
-        $text .= "NDATA $ent->{ndata}"        if( $ent->{ndata});
+    elsif( $ent->{'sysid'})
+      { my $text= "<!ENTITY $ent->{'name'} ";
+        $text .= "SYSTEM \"$ent->{'sysid'}\" " if( $ent->{'sysid'});
+        $text .= "PUBLIC \"$ent->{'pubid'}\" " if( $ent->{'pubid'});
+        $text .= "NDATA $ent->{'ndata'}"        if( $ent->{'ndata'});
         $text .= ">";
         return $text;
       }
@@ -375,7 +656,7 @@ sub gi { return $index2gi[$_[0]->{gi}]; }
 sub closed 
   { my $elt= shift;
     my $t= $elt->twig || return;
-    my $curr_elt= $t->{twig_elt};
+    my $curr_elt= $t->{twig_current};
     return unless( $curr_elt);
     return $curr_elt->in( $elt);
   }
@@ -464,10 +745,23 @@ sub del_att { delete $_[0]->{'att'}->{$_[1]}; }
 
 # get or set the id attribute
 sub set_id 
-  { $_[0]->set_att($ID, $_[1]); 
-}
+  { my( $elt, $id)= @_;
+    $elt->set_att($ID, $_[1]); 
+    my $t= $elt->twig || return;
+    $elt->twig->{twig_id_list}->{$id}= $elt;
+  }
+
 sub id { return $_[0]->{'att'}->{$ID}; }
 
+# delete the id attribute and remove the element from the id list
+sub del_id 
+  { my $elt= shift;
+    my $id= $elt->{'att'}->{$ID} or return;
+    my $t= $elt->twig && delete $elt->twig->{twig_id_list}->{$id};
+    delete $elt->{'att'}->{$ID}; 
+  }
+
+
 sub children
   { my $elt= shift;
     my $gi= shift;
@@ -553,10 +847,14 @@ sub in($$)
     return 0;           
   }
 
+# returns the depth level of the element
+# if 2 parameter are used then counts the 2cd element name in the
+# ancestors list
 sub level
   { my $elt= shift;
     my $level=0;
-    while( $elt= $elt->parent) { $level++; }
+    my $name=shift || '';
+    while( $elt= $elt->parent) { $level++ if( !$name || ($name eq $elt->gi)); }
     return $level;           
   }
 
@@ -650,13 +948,17 @@ sub erase
             $elt->parent->set_last_child( $elt->next_sibling);
           }
       }
+    # elt is not referenced any more, so it will be DESTROYed
+    # so we'd better break the links to its children
+    undef $elt->{'first_child'};
+    undef $elt->{'last_child'};
+
   }
         
 
 
 # paste elt somewhere around ref
-# if the second argument is ommited then 'first_child' is used
-# first_child (default), last_child, before, after
+# pos can be first_child (default), last_child, before, after
 sub paste
   { my $elt= shift;
     my $pos;
@@ -731,6 +1033,32 @@ sub paste
       { croak "tried to paste in wrong position ($pos)\n"; }
   }
 
+
+# recursively copy an element and returns the copy (can be huge and long)
+sub copy
+  { my $elt= shift;
+    my $copy= new XML::Twig::Elt;
+    $copy->set_gi( $elt->gi);
+    if( my $atts= $elt->atts)
+      { my %atts= %{$atts}; # we want to do a real copy of the attributes
+        $copy->set_atts( \%atts);
+      }
+
+    if( $elt->gi eq PCDATA)
+      { $copy->set_pcdata( $elt->pcdata);
+        return $copy;
+      }
+    else
+      { my @children= $elt->children;
+        foreach my $child (@children)
+          { my $child_copy= $child->copy;
+            $child_copy->paste( 'last_child', $copy);
+          }
+      }
+    return $copy;
+  }
+
+
 sub delete
   { my $elt= shift;
     $elt->cut;
@@ -740,14 +1068,17 @@ sub delete
 sub DESTROY
   { my $elt= shift;
 
-    foreach( $elt->children) { $_->DESTROY; }
+    foreach( @{[$elt->children]}) { $_->DESTROY; }
     # destroy all references in the tree
-    $elt->set_parent( undef);
-    $elt->set_first_child( undef);
-    $elt->set_last_child( undef);
-    $elt->set_prev_sibling( undef);
-    $elt->set_next_sibling( undef);
-
+    $elt->{'parent'}      = undef; # $elt->set_parent( undef);
+    $elt->{'first_child'} = undef; # $elt->set_first_child( undef);
+    $elt->{'last_child'}  = undef; # $elt->set_last_child( undef);
+    $elt->{'prev_sibling'}= undef; # $elt->set_prev_sibling( undef);
+    $elt->{'next_sibling'}= undef; # $elt->set_next_sibling( undef);
+    # the id reference also needs to be destroyed
+    $elt->del_id;
+    $elt->{'att'}= undef;         # $elt->set_atts( undef);
+    $elt= undef;
   }
 
 sub start_tag
@@ -818,14 +1149,22 @@ sub flush
     print $fh $elt->end_tag;
   }
 
+
 # same as print but except... it does not print but rather returns the string
+# if the second parameter is set then only the content is returned, not the
+# start and end tags of the element (but the tags of the included elements are
+# returned)
 sub sprint
-  { my ($elt)= @_;
+  { my $elt= shift;
+    my $no_tag= shift || 0;
+
     my $gi= $elt->gi;
 
     if( $gi eq PCDATA) { return $elt->pcdata; }
 
-    my $string= $elt->start_tag;
+    my $string='';
+
+    $string= $elt->start_tag unless( $no_tag);
 
     # sprint the children
     my $child= $elt->first_child;
@@ -833,9 +1172,11 @@ sub sprint
       { $string.= $child->sprint;
         $child= $child->next_sibling;
       }
-    return $string.$elt->end_tag;
+    $string .= $elt->end_tag unless( $no_tag);
+    return $string;
   }
 
+
 # returns just the text, no tags, for an element
 sub text
   { my $elt= shift;
@@ -845,13 +1186,67 @@ sub text
 
     my $child= $elt->first_child;
     while( $child)
-      { $string.= $child->text || '';
+      { $string.= defined($child->text) ? $child->text : '';
         $child= $child->next_sibling;
       }
     return $string;
   }
 
 
+# creates a single pcdata element containing the text as child of the element
+sub set_text
+  { my $elt= shift;
+    my $string= shift;
+
+    if( $elt->gi eq PCDATA) { return $elt->set_pcdata( $string); }
+
+    foreach my $child (@{[$elt->children]})
+      { $child->cut; }
+
+    my $pcdata= new XML::Twig::Elt;
+    $pcdata->set_gi( PCDATA);
+    $pcdata->set_pcdata( $string);
+    $pcdata->paste( $elt);
+
+    return;
+  }
+
+# set the content of an element from a list of strings and elements
+sub set_content
+  { my $elt= shift;
+
+    foreach my $child (@{[$elt->children]})
+      { $child->cut; }
+
+    foreach my $child (@_)
+      { if( ref $child)
+          { $child->paste( 'last_child', $elt); }
+        else
+          { my $pcdata= new XML::Twig::Elt;
+            $pcdata->set_gi( PCDATA);
+            $pcdata->set_pcdata( $child);
+            $pcdata->paste( 'last_child', $elt);  
+          }
+      }
+
+    return;
+  }
+
+# inserts an element (whose gi is given) as child of the element
+# all children of the element are now children of the new element
+# returns the new element
+sub insert
+  { my ($elt, $gi)= @_;
+    my $new_elt= new XML::Twig::Elt;
+    $new_elt->set_gi( $gi);
+    foreach my $child (@{[$elt->children]})
+      { $child->cut;
+        $child->paste( 'last_child', $new_elt);
+      }
+    $new_elt->paste( $elt);
+    return $new_elt;
+  }
+
 
 __END__
 
@@ -867,7 +1262,7 @@ XML::Twig - A perl module for processing huge XML documents in tree mode.
         $t->print;
 
     chunk mode 
-        my $t= new XML::Twig( Handlers => { section => \&flush});
+        my $t= new XML::Twig( TwigHandlers => { section => \&flush});
         $t->parsefile( 'doc.xml');
         $t->flush;
         sub flush { $_[0]->flush; }
@@ -893,7 +1288,7 @@ see C<L</BUGS>>
 
 =over 4
 
-=item - new 
+=item new 
 
 This is a class method, the constructor for XML::Twig. Options are passed
 as keyword value pairs. Recognized options are the same as L<XML::Parser>,
@@ -907,11 +1302,39 @@ This argument replaces the corresponding XML::Parser argument. It consists
 of a hash { gi => \&handler} 
 A gi (generic identifier I guess) is just a tag name by the way.
 When an element is CLOSED the corresponding handler is called, with 2 arguments,
-the twig and the C<L</Element>>. The twig includes the document tree taht has been built so far, the element is the complete sub-tree for the element.
+the twig and the C<L</Element>>. The twig includes the document tree taht has been 
+built so far, the element is the complete sub-tree for the element.
 Text is stored in elements which gi is #PCDATA (due to mixed content, text
 and sub-element in an element there is no way to store the text as just an
 attribute of the enclosing element).
 
+=item LoadDTD
+
+If this argument is set to a true value, parse or parsefile on the twig will load  
+the DTD information. This information can then be accessed through the twig, 
+in a DTDHandler for example. This will load even an external DTD.
+
+See L<DTD Handling> for more information
+
+=item DTDHandler
+
+Sets a handler that will be called once the doctype (and the DTD) have been loaded,
+with 2 arguments, the twig and the DTD.
+
+-item StartTagHandlers
+
+A hash { gi => \&handler}. Sets element handlers that are called when the element 
+is open (at the end of the XML::Parser Start handler). THe handlers are called with
+2 params: the twig and the element. The element is empty at that point, its attributes
+are created though.
+
+The main use for those handlers is probably to create temporary attributes that will
+be used when processing the element with the normal TwigHanlder.  
+
+-item CharHandler
+
+A reference to a subroutine that will be called every time PCDATA.
+
 =item - Id
 
 This optional argument gives the name of an attribute that can be used as
@@ -936,7 +1359,7 @@ Returns the entity list of a twig
 Performs a (very fast) global change. All elements old_gi are now new_gi.
 See C<L</BUGS>>
 
-=item flush            OPTIONAL_FILEHANDLE
+=item flush            OPTIONAL_FILEHANDLE OPTIONNAL_OPTIONS
 
 Flushes a twig up to (and including) the current element, then deletes
 all unnecessary elements from the tree that's kept in memory.
@@ -945,20 +1368,39 @@ flush from handlers you don't have to worry about anything. Just keep
 flushing the twig every time you're done with a sub-tree and it will
 come out well-formed. After the whole parsing don't forget to flush
 one more time to print the end of the document.
-The doctype and entity declarations are also printed by the way.
+The doctype and entity declarations are also printed.
+
+OPTIONNAL_OPTIONS
+
+=over 4
+
+= item Update_DTD
+
+Use that option if you have updated the (internal) DTD and/or the enity list
+and you want the updated DTD to be output 
+
+Example $t->flush( Update_DTD => 1);
+        $t->flush( \*FILE, Update_DTD => 1);
+        $t->flush( \*FILE);
+
+=back
+
 
 flush take an optional filehandle as an argument.
 
-=item print            OPTIONNAL_FILEHANDLE
+=item print            OPTIONNAL_FILEHANDLE OPTIONNAL_OPTIONS
 
-Oddly enough this is just an alias for the flush method. For sanity's sake
-just try to use it only for trees that have been completely loaded, there
-is no check right now but I might add one soon.
+Prints the whole document associated with the twig. To be used only AFTER the
+parse.
 
-=item print_prolog     OPTIONNAL_FILEHANDLE
+OPTIONNAL_OPTIONS: see L<flush>.
+
+
+=item print_prolog     OPTIONNAL_FILEHANDLE OPTIONNAL_OPTIONS
 
 Prints the prolog (XML declaration + DTD + entity declarations) of a document.
-Does not print the DTD at the moment.
+
+OPTIONNAL_OPTIONS: see L<flush>.
 
 =back 
 
@@ -1026,7 +1468,7 @@ Returns the next sibling of the element, or the first one whose gi is $gi.
 
 =item atts
 
-Returns a hash containing the element attributes
+Returns a hash ref containing the element attributes
 
 =item set_atts      ({att1=>$att1_val, att2=> $att2_val... )
 
@@ -1057,6 +1499,11 @@ See C<L</elt_id>> to change the id attribute name
 
 Gets the id attribute vakue
 
+=item del_id       ($id)
+
+Deletes the id attribute of the element and remove it from the id list
+for the document
+
 =item children     ($optional_gi)
 
 Returns the list of children (optionally whose gi is $gi) of the element
@@ -1079,9 +1526,10 @@ Returns the previous elt (optionally whose gi is $gi) of the element. This
 is the first element which open the current one. So it's usually either
 the last descendant of the previous sibling or simply the parent
 
-=item level
+=item level       ($optionnal_gi)
 
 Returns the depth of the element in the tree (root is 1)
+If the optionnal gi is given then only ancestors of the given type are counted 
 
 =item in           ($potential_parent)
 
@@ -1158,6 +1606,23 @@ with caution!
 Returns a string consisting of all the PCDATA in an element, without the
 tagging
 
+=item set_text        ($string)
+
+Sets the text for the element: if the element is a PCDATA, just set its
+text, otherwise cut all the children of the element and create a single
+PCDATA child for it, which holds the text
+
+=item set_content    (@list_of_elt_and_strings)
+
+Sets the content for the element, from as list of strings and elements.
+Cuts all the element children, then pastes the list elements, creating a 
+PCDATA element for strings.
+
+=item insert         ($gi)
+
+Inserts an element $gi as the only child of the element, all children of 
+the element are set as children of the new element, returns the new element
+
 =item private methods
 
 =over 4
@@ -1181,14 +1646,14 @@ tagging
 =back
 
 Those methods should not be used, unless of course you find some creative 
-and interesting ways to do it.
+and interesting, not to mention usefull, ways to do it.
 
 
 =back
 
 =head2 Entity_list
 
-=over4
+=over 4
 
 =item new
 
@@ -1229,10 +1694,64 @@ Returns the entity declaration text
 
 =head1 EXAMPLES
 
-See the test file in XML-Twig-1.6/t/test1.t
+See the test file in XML-Twig-1.6/t/test[1-n].t 
+
+To figure out what flush does call the following script with an xml file and an element name as arguments
+
+use XML::Twig;
+
+my ($file, $elt)= @ARGV;
+my $t= new XML::Twig( TwigHandlers => 
+    { $elt => sub {$_[0]->flush; print "\n[flushed here]\n";} });
+$t->parsefile( $file, ErrorContext => 2);
+$t->flush;
+print "\n";
 
 
-=head1 HINTS
+
+=head1 NOTES
+
+=head2 DTD Handling
+
+3 possibilities here
+
+=over 4
+
+=item No DTD
+
+No doctype,  no DTD information, no entitiy information, the world is simple...
+
+=item Internal DTD
+
+The XML document includes an internal DTD, and maybe entity declarations
+
+If you use the LoadDTD option when creating the twig the DTD information and the entity 
+declarations can be accessed. 
+
+The DTD and the entity declarations will be flush'ed (or print'ed) either asis
+(if they have not been modified) or as reconstructed (poorly, comments are lost, 
+order is not kept, due to it's content this DTD should not be viewed bu anyone) 
+if they have been modified. You can also modify them directly by changing the 
+$twig->{twig_doctype}->{internal} field (straight from XML::Parser, see the Doctype 
+handler doc)
+
+=item External DTD
+
+The XML document includes a reference to an external DTD, and maybe entity declarations.
+
+If you use the LoadDTD when creating the twig the DTD information and the entity 
+declarations can be accessed. The entity declarations will be flush'ed (or print'ed)
+either asis (if they have not been modified) or as reconstructed (badly, comments are
+lost, order is not kept).
+
+You can change the doctype through the $twig->set_doctype method and print the dtd 
+through the $twig->dtd_text or $twig->dtd_print methods.
+
+If you need to modify the entity list this is probably the easiest way to do it.
+
+=back
+
+=head2 Flush
 
 If you set handlers and use flush, do not forget to flush the twig one
 last time AFTER the parsing, or you might be missing the end of the document.
@@ -1242,6 +1761,8 @@ if you have handlers for nested elements the inner handlers will be called
 first. It makes it for example trickier than it would seem to number nested
 clauses.
 
+
+
 =head1 BUGS
 
 =over 4
@@ -1262,9 +1783,6 @@ Does not work if you do:
 
 XML::Twig should really prevent calls to some XML::Parser methods, especially the setHandlers one.
 
-=item - Notation declarations
-
-Are not output (in fact they are completely ignored).
 
 =back
 
@@ -1290,10 +1808,15 @@ tags, like adding an ID, or taking care of the autonumbering.
 
 Next version...
 
-=item - create an element (not a twig) from a string
 
 =back
 
+
+=head1 BENCHMARKS
+
+You can use the C<benchmark> file to do additional bechmarks.
+Please send me bechmark information for additional systems.
+
 =head1 AUTHOR
 
 Michel Rodriguez <m.v.rodriguez@ieee.org>
